package com.vince.digikamdbvalidator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URLDecoder;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;
import java.util.TreeMap;

import org.sqlite.SQLiteConfig;

import com.vince.digikamdbvalidator.util.AlbumRoot;
import com.vince.digikamdbvalidator.util.Tree;

public class DigikamDBValidator
{
	private final static String			SQL_CHECK_ALBUM_ROOT			= "SELECT id, label, type, identifier, specificPath FROM AlbumRoots";
	private final static String			SQL_CHECK_ALBUMS1				= "SELECT Albums.id FROM Albums LEFT OUTER JOIN AlbumRoots ON Albums.albumRoot=AlbumRoots.id WHERE AlbumRoots.id IS NULL";
	private final static String			SQL_CHECK_ALBUMS2				= "SELECT Albums.id , Albums.relativePath, AlbumRoots.id as rootid FROM Albums LEFT OUTER JOIN AlbumRoots ON Albums.albumRoot=AlbumRoots.id WHERE Albums.id IN (SELECT Albums.id FROM Albums LEFT OUTER JOIN AlbumRoots ON Albums.albumRoot=AlbumRoots.id WHERE AlbumRoots.id IS NOT NULL)";
	private final static String			SQL_CHECK_IMAGECOMMENTS			= "SELECT imageComments.id FROM imageComments LEFT OUTER JOIN Images ON Images.id=ImageComments.imageid WHERE Images.name IS NULL";
	private final static String			SQL_CHECK_IMAGECOPYRIGHT		= "SELECT imageCopyright.id FROM imageCopyright LEFT OUTER JOIN Images ON Images.id=ImageCopyright.imageid WHERE Images.name IS NULL";
	private final static String			SQL_CHECK_IMAGEHAARMATRIX		= "SELECT imageHaarMatrix.imageid FROM ImageHaarMatrix LEFT OUTER JOIN Images ON Images.id=ImageHaarMatrix.imageid WHERE Images.name IS NULL";
	private final static String			SQL_CHECK_IMAGEINFORMATION		= "SELECT ImageInformation.imageid FROM ImageInformation LEFT OUTER JOIN Images ON Images.id=ImageInformation.imageid WHERE Images.name IS NULL";
	private final static String			SQL_CHECK_IMAGEINFORMATIONCOUNT	= "SELECT (SELECT COUNT(*) FROM ImageInformation) AS count1, (SELECT COUNT(*) FROM Images) AS count2";
	private final static String			SQL_CHECK_IMAGEMETADATA			= "SELECT ImageMetadata.imageid FROM ImageMetadata LEFT OUTER JOIN Images ON Images.id=ImageMetadata.imageid WHERE Images.name IS NULL";
	private final static String			SQL_CHECK_IMAGEPOSITIONS		= "SELECT ImagePositions.imageid FROM ImagePositions LEFT OUTER JOIN Images ON Images.id=ImagePositions.imageid WHERE Images.name IS NULL";
	private final static String			SQL_CHECK_IMAGEPROPERTIES		= "SELECT ImageProperties.rowid FROM ImageProperties LEFT OUTER JOIN Images ON Images.id=ImageProperties.imageid WHERE Images.name IS NULL";
	private final static String			SQL_CHECK_IMAGETAGS1			= "SELECT ImageTags.rowid FROM ImageTags LEFT OUTER JOIN Images ON Images.id=ImageTags.imageid WHERE Images.name IS NULL";
	private final static String			SQL_CHECK_IMAGETAGS2			= "SELECT ImageTags.rowid FROM ImageTags LEFT OUTER JOIN Tags ON Tags.id=ImageTags.tagid WHERE Tags.name IS NULL";
	private final static String			SQL_CHECK_IMAGES1				= "SELECT Images.id, Images.album FROM Images LEFT OUTER JOIN Albums ON Albums.id=Images.album WHERE Albums.id IS NULL AND Images.album IS NOT NULL AND Images.status!=3";
	private final static String			SQL_CHECK_IMAGES2				= "SELECT Images.id, Images.name, Albums.relativePath, AlbumRoots.id as rootid FROM Images, Albums, AlbumRoots WHERE Albums.id=Images.album AND AlbumRoots.id=Albums.albumRoot AND Images.id IN (SELECT Images.id FROM Images LEFT OUTER JOIN Albums ON Albums.id=Images.album WHERE Albums.id IS NOT NULL)";
	private final static String			SQL_CHECK_IMAGES3				= "SELECT album FROM Images WHERE name IN (SELECT name FROM Images WHERE id=?) AND modificationDate IN (SELECT modificationDate FROM Images WHERE id=?) AND album NOT IN (SELECT album FROM Images WHERE id=?)";
	private final static String			SQL_CHECK_TAGS					= "SELECT T1.id FROM Tags AS T1 LEFT OUTER JOIN Tags As T2 ON T2.id=T1.pid WHERE T1.pid != 0 AND T2.id IS NULL";
	private final static String			SQL_LIST_TAGS					= "SELECT id FROM Tags WHERE pid=?";
	private final static String			SQL_CHECK_TAGSTREE1				= "SELECT COUNT(*) AS count FROM TagsTree WHERE pid=? AND id=?";
	private final static String			SQL_CHECK_TAGSTREE2				= "SELECT TagsTree.rowid, TagsTree.id, TagsTree.pid FROM TagsTree";
	private final static String			SQL_GET_DB_VERSION				= "SELECT value FROM Settings WHERE keyword=\"DBVersion\"";
	private final static int			SUPPORTED_DB_VERSION			= 5;
	private final static String			NOTFOUND						= "DIGIKAMDBVALIDATORNOTFOUND";
	private final static String			NOTMOUNTED						= "DIGIKAMDBVALIDATORNOTMOUNTED";

	private final static String			VERSION							= "V1.2 - Dec 1 2011";

	private Connection					conn;
	private Tree<Integer>				tags							= new Tree<Integer>(Integer.valueOf(0));
	private TreeMap<Integer, AlbumRoot>	albumRoots						= new TreeMap<Integer, AlbumRoot>();

	public DigikamDBValidator(String db)
	{
		try
		{
			Class.forName("org.sqlite.JDBC");
		}
		catch(ClassNotFoundException e)
		{
			System.err.println("SQLite JDBC driver not found ...");
		}
		try
		{
			SQLiteConfig cfg = new SQLiteConfig();
			cfg.setReadOnly(true);
			conn = DriverManager.getConnection("jdbc:sqlite:" + db, cfg.toProperties());
			// System.out.println(String.format("SQLite driver running in %s mode",
			// SQLiteJDBCLoader.isNativeMode() ? "native" : "pure-java"));
			PreparedStatement prep = conn.prepareStatement(SQL_GET_DB_VERSION);
			ResultSet rs = prep.executeQuery();
			rs.next();
			String dbVersion = rs.getString("value");
			System.out.print("Database Schema version is " + dbVersion + " - ");
			if(Integer.valueOf(dbVersion).intValue() > SUPPORTED_DB_VERSION)
			{
				System.out.println("This is higher than currently supported version of " + SUPPORTED_DB_VERSION + " ... Not all tables will be validated\n");
			}
			else
			{
				System.out.println("All tables will be validated\n");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
	}

	@Override
	protected void finalize() throws Throwable
	{
		conn.close();
		super.finalize();
	}

	public int validateAlbumRoots()
	{
		int errorCount = 0;
		boolean skipped = false;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_ALBUM_ROOT);
			ResultSet rs = prep.executeQuery();
			int albumType;
			int albumId;
			String albumIdentifier;
			String albumLabel;
			String albumspecificPath;
			String albumPath;
			System.out.print("Checking AlbumRoots...");
			while(rs.next())
			{
				albumId = rs.getInt("id");
				albumLabel = rs.getString("label");
				albumType = rs.getInt("type");
				albumIdentifier = rs.getString("identifier");
				albumspecificPath = rs.getString("specificPath");
				albumPath = getAlbumPath(albumType, albumIdentifier, albumspecificPath);
				albumRoots.put(Integer.valueOf(albumId), new AlbumRoot(albumId, albumLabel, albumPath));
				if(albumPath.equals(NOTFOUND))
				{
					if(errorCount == 0 && !skipped)
					{
						System.out.println();
					}
					System.out.println("\tAlbumRoot " + albumId + " is on a device which is not currently connected\t\t");
					skipped = true;
				}
				else if(albumPath.equals(NOTMOUNTED))
				{
					if(errorCount == 0 && !skipped)
					{
						System.out.println();
					}
					System.out.println("\tAlbumRoot " + albumId + " is on a device which is currently not mounted\t\t");
					skipped = true;
				}
				else
				{
					File f = new File(albumPath);
					if(!f.exists())
					{
						if(errorCount == 0 && !skipped)
						{
							System.out.println();
						}
						System.out.println("\tAlbumRoot " + albumId + " doesn't exist\t\t" + albumPath);
						errorCount++;
					}
					else if(!f.isDirectory())
					{
						if(errorCount == 0 && !skipped)
						{
							System.out.println();
						}
						System.out.println("\tAlbumRoot " + albumId + " does not point to a directory\t\t" + albumPath);
						errorCount++;
					}
				}
			}
			rs.close();
			prep.close();
			if(errorCount == 0 && !skipped)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateAlbums()
	{
		int errorCount = 0;
		boolean skipped = false;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_ALBUMS1);
			ResultSet rs = prep.executeQuery();
			System.out.print("Checking Albums...");
			int albumId;
			while(rs.next())
			{
				albumId = rs.getInt("id");
				if(errorCount == 0 && !skipped)
				{
					System.out.println();
				}
				System.out.println("\tAlbum " + albumId + " points to inexistiant AlbumRoot");
				errorCount++;
			}
			rs.close();
			prep.close();
			
			int rootid;
			String albumPath;
			String albumRootPath;
			prep = conn.prepareStatement(SQL_CHECK_ALBUMS2);
			rs = prep.executeQuery();
			while(rs.next())
			{
				albumId = rs.getInt("id");
				rootid = rs.getInt("rootid");
				albumRootPath = albumRoots.get(Integer.valueOf(rootid)).getPath();
				albumPath = albumRootPath + rs.getString("relativePath");
				if(albumRootPath.equals(NOTFOUND))
				{
					if(errorCount == 0 && !skipped)
					{
						System.out.println();
					}
					System.out.println("\tAlbum " + albumId + " is on a device which is not currently connected\t\t");
					skipped = true;
				}
				else if(albumRootPath.equals(NOTMOUNTED))
				{
					if(errorCount == 0 && !skipped)
					{
						System.out.println();
					}
					System.out.println("\tAlbum " + albumId + " is on a device which is currently not mounted\t\t");
					skipped = true;
				}
				else
				{
					File f = new File(albumPath);
					if(!f.exists())
					{
						if(errorCount == 0 && !skipped)
						{
							System.out.println();
						}
						System.out.println("\tAlbum " + albumId + " doesn't exist\t\t" + albumPath);
						errorCount++;
					}
					else if(!f.isDirectory())
					{
						if(errorCount == 0 && !skipped)
						{
							System.out.println();
						}
						System.out.println("\tAlbum " + albumId + " does not point to a directory\t\t" + albumPath);
						errorCount++;
					}
				}
			}
			rs.close();
			prep.close();
			if(errorCount == 0 && !skipped)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;

	}

	public int validateImageComments()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGECOMMENTS);
			ResultSet rs = prep.executeQuery();
			int commentId;
			System.out.print("Checking Image Comments...");
			while(rs.next())
			{
				commentId = rs.getInt("id");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImageComment " + commentId + " points to inexistiant image");
				errorCount++;
			}
			rs.close();
			prep.close();
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateImageCopyright()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGECOPYRIGHT);
			ResultSet rs = prep.executeQuery();
			int commentId;
			System.out.print("Checking Image Copyrights...");
			while(rs.next())
			{
				commentId = rs.getInt("id");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImageCopyright " + commentId + " points to inexistiant image");
				errorCount++;
			}
			rs.close();
			prep.close();
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateImageHaarMatrix()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGEHAARMATRIX);
			ResultSet rs = prep.executeQuery();
			int hashId;
			System.out.print("Checking Image Hashes...");
			while(rs.next())
			{
				hashId = rs.getInt("imageid");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImage Hash " + hashId + " points to inexistiant image");
				errorCount++;
			}
			rs.close();
			prep.close();
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateImageInformation()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGEINFORMATION);
			ResultSet rs = prep.executeQuery();
			int id;
			System.out.print("Checking Image Informations...");
			while(rs.next())
			{
				id = rs.getInt("imageid");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImage Information " + id + " points to inexistiant image");
				errorCount++;
			}
			rs.close();
			prep.close();
			prep = conn.prepareStatement(SQL_CHECK_IMAGEINFORMATIONCOUNT);
			rs = prep.executeQuery();
			rs.next();
			int countInfo = rs.getInt("count1");
			int countImages = rs.getInt("count2");
			rs.close();
			prep.close();
			if(countImages > countInfo)
			{
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\t" + (countImages - countInfo) + " images don't have information registered");
				errorCount += (countImages - countInfo);
			}
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateImageMetadata()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGEMETADATA);
			ResultSet rs = prep.executeQuery();
			int id;
			System.out.print("Checking Image Metadata...");
			while(rs.next())
			{
				id = rs.getInt("imageid");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImage Metadata " + id + " points to inexistiant image");
				errorCount++;
			}
			rs.close();
			prep.close();
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateImagePositions()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGEPOSITIONS);
			ResultSet rs = prep.executeQuery();
			int id;
			System.out.print("Checking Image Positions...");
			while(rs.next())
			{
				id = rs.getInt("imageid");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImage Positions " + id + " points to inexistiant image");
				errorCount++;
			}
			rs.close();
			prep.close();
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateImageProperties()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGEPROPERTIES);
			ResultSet rs = prep.executeQuery();
			int rowId;
			System.out.print("Checking Image Properties...");
			while(rs.next())
			{
				rowId = rs.getInt("rowid");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImage Property " + rowId + " points to inexistiant image");
				errorCount++;
			}
			rs.close();
			prep.close();
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateImageTags()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGETAGS1);
			ResultSet rs = prep.executeQuery();
			int rowId;
			System.out.print("Checking Image Tags...");
			while(rs.next())
			{
				rowId = rs.getInt("rowid");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImage Tag " + rowId + " points to inexistiant image");
				errorCount++;
			}
			rs.close();
			prep.close();
			prep = conn.prepareStatement(SQL_CHECK_IMAGETAGS2);
			rs = prep.executeQuery();
			while(rs.next())
			{
				rowId = rs.getInt("rowid");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tImage Tag " + rowId + " points to inexistiant Tag");
				errorCount++;
			}
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateImages()
	{
		int errorCount = 0;
		boolean skipped = false;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_IMAGES1);
			ResultSet rs = prep.executeQuery();
			int id;
			int imageAlbum;
			System.out.print("Checking Images...");
			while(rs.next())
			{
				id = rs.getInt("id");
				imageAlbum = rs.getInt("album");
				if(errorCount == 0 && !skipped)
				{
					System.out.println();
				}
				errorCount++;
				PreparedStatement prep2 = conn.prepareStatement(SQL_CHECK_IMAGES3);
				prep2.setInt(1, id);
				prep2.setInt(2, id);
				prep2.setInt(3, id);
				ResultSet rs2 = prep2.executeQuery();
				if(rs2.next())
				{
					int album = rs2.getInt("album");
					System.out.println("\tImage " + id + " points to inexistiant Album " + imageAlbum + " but similar image also found in album " + album);
				}
				else
				{
					System.out.println("\tImage " + id + " points to inexistiant Album " + imageAlbum);
				}
				rs2.close();
				prep2.close();
			}
			rs.close();
			prep.close();
			prep = conn.prepareStatement(SQL_CHECK_IMAGES2);
			rs = prep.executeQuery();
			int rootid;
			String imagePath;
			String albumRootPath;
			File f;
			while(rs.next())
			{
				id = rs.getInt("id");
				rootid = rs.getInt("rootid");
				albumRootPath = albumRoots.get(Integer.valueOf(rootid)).getPath();
				imagePath = albumRootPath + rs.getString("relativePath") + "/" + rs.getString("name");
				if(albumRootPath.equals(NOTFOUND))
				{
					if(errorCount == 0 && !skipped)
					{
						System.out.println();
					}
					System.out.println("\tImage " + id + " is on a device which is not currently connected\t\t");
					skipped = true;
				}
				else if(albumRootPath.equals(NOTMOUNTED))
				{
					if(errorCount == 0 && !skipped)
					{
						System.out.println();
					}
					System.out.println("\tImage " + id + " is on a device which is currently not mounted\t\t");
					skipped = true;
				}
				else
				{
					f = new File(imagePath);
					if(!f.exists())
					{
						if(errorCount == 0 && !skipped)
						{
							System.out.println();
						}
						System.out.println("\tImage " + id + " doesn't exist\t\t" + imagePath);
						errorCount++;
					}
					else if(!f.isFile())
					{
						if(errorCount == 0 && !skipped)
						{
							System.out.println();
						}
						System.out.println("\tImage " + id + " does not point to a file\t\t" + imagePath);
						errorCount++;
					}
				}
			}
			rs.close();
			prep.close();
			if(errorCount == 0 && !skipped)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateTags()
	{
		int errorCount = 0;
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_TAGS);
			ResultSet rs = prep.executeQuery();
			int id;
			System.out.print("Checking Tags...");
			while(rs.next())
			{
				id = rs.getInt("id");
				if(errorCount == 0)
				{
					System.out.println();
				}
				System.out.println("\tTag " + id + " has invalid parent Tag");
				errorCount++;
			}
			rs.close();
			prep.close();
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	public int validateTagsTree()
	{
		int errorCount = 0;
		System.out.print("Building Tag Tree...");
		addTags(0);
		System.out.println("OK");
		System.out.print("Checking TagsTree for completeness...");
		errorCount += processLeafs(tags);
		if(errorCount == 0)
		{
			System.out.println("OK");
		}
		System.out.print("Checking TagsTree for orphaned entries...");
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_CHECK_TAGSTREE2);
			ResultSet rs = prep.executeQuery();
			int rowid;
			int id;
			int pid;
			while(rs.next())
			{
				rowid = rs.getInt("rowid");
				id = rs.getInt("id");
				pid = rs.getInt("pid");
				Tree<Integer> t = tags.getTree(Integer.valueOf(pid));
				if(t == null)
				{
					if(errorCount == 0)
					{
						System.out.println();
					}
					System.out.println("\tEntry with rowid " + rowid + " refers to missing pid\t\t" + pid);
					errorCount++;
				}
				else if(!t.getAllChildren().contains(Integer.valueOf(id)))
				{
					if(errorCount == 0)
					{
						System.out.println();
					}
					System.out.println("\tEntry with rowid " + rowid + " points to incorrect id\t\tid=" + id + " pid=" + pid);
					// System.out.println(t.printAllChildren());
					errorCount++;
				}
			}
			rs.close();
			prep.close();
			if(errorCount == 0)
			{
				System.out.println("OK");
			}
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		return errorCount;
	}

	private void addTags(int pid)
	{
		try
		{
			PreparedStatement prep = conn.prepareStatement(SQL_LIST_TAGS);
			prep.setInt(1, pid);
			ResultSet rs = prep.executeQuery();
			int id;
			while(rs.next())
			{
				id = rs.getInt("id");
				tags.addLeaf(Integer.valueOf(pid), Integer.valueOf(id));
				addTags(id);
			}
			rs.close();
			prep.close();
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
	}

	private int processLeafs(Tree<Integer> t)
	{
		int errorCount = 0;
		Collection<Integer> children = t.getAllChildren();
		Iterator<Integer> it = children.iterator();
		while(it.hasNext())
		{
			Integer child = it.next();
			try
			{
				PreparedStatement prep = conn.prepareStatement(SQL_CHECK_TAGSTREE1);
				prep.setInt(1, t.getHead().intValue());
				prep.setInt(2, child.intValue());
				ResultSet rs = prep.executeQuery();
				rs.next();
				int count = rs.getInt("count");
				rs.close();
				prep.close();
				if(count != 1)
				{
					if(errorCount == 0)
					{
						System.out.println();
					}
					System.out.println("\tMissing TagsTree entry for id=" + t.getHead() + " and pid=" + child);
					errorCount++;
				}
			}
			catch(SQLException e)
			{
				e.printStackTrace();
			}
		}
		for(Tree<Integer> subchild : t.getSubTrees())
		{
			errorCount += processLeafs(subchild);
		}
		return errorCount;
	}

	private String getAlbumPath(int albumType, String albumIdentifier, String albumspecificPath)
	{
		String albumPath = "";
		if(albumspecificPath == null)
		{
			albumspecificPath = "";
		}
		if(albumType == 1 || albumType == 2)
		{
			String uuid = albumIdentifier.split("uuid=")[1];
			File device = new File("/dev/disk/by-uuid/" + uuid);
			try
			{
				if(device.exists())
				{
					String deviceName = device.getCanonicalPath();
					BufferedReader input = new BufferedReader(new FileReader("/etc/mtab"));
					String line;
					String mountPoint = "";
					boolean found = false;
					while((line = input.readLine()) != null && !found)
					{
						if(line.contains(deviceName))
						{
							mountPoint = line.split(" ")[1];
							found = true;
						}
						else
						{
							device = new File(line.split(" ")[0]);
							if(device.exists())
							{
								String deviceName2 = device.getCanonicalPath();
								if(deviceName.equals(deviceName2))
								{
									mountPoint = line.split(" ")[1];
									found = true;
								}
							}
						}
					}
					if(found)
					{
						albumPath = mountPoint + albumspecificPath;
					}
					else
					{
						albumPath = NOTMOUNTED;
					}
				}
				else
				{
					albumPath = NOTFOUND;
				}
			}
			catch(IOException e1)
			{
				albumPath = NOTFOUND;
			}
		}
		else if(albumType == 3)
		{
			try
			{
				albumPath = URLDecoder.decode(albumIdentifier.split("mountpath=")[1], "UTF-8") + albumspecificPath;
				if(albumPath.charAt(albumPath.length()-1) == '/')
				{
					albumPath = albumPath.substring(0, albumPath.length()-1);
				}
				BufferedReader input = new BufferedReader(new FileReader("/etc/mtab"));
				String line;
				boolean found = false;
				while((line = input.readLine()) != null && !found)
				{
					if(line.contains(albumPath))
					{
						found = true;
					}
				}
				if(!found)
				{
					albumPath = NOTMOUNTED;
				}
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
		}
		return albumPath.replace("//", "/");
	}

	/**
	 * @param args
	 */
	public static void main(String[] args)
	{
		String kdeLocalPrefix = null;
		System.out.println("Digikam DB Validator " + VERSION);
		System.out.println("Report bugs to <photo@tassy.net>\n");
		if(!"Linux".equals(System.getProperty("os.name")))
		{
			System.err.println("Digikam DB Validator is only supported on Linux for now");
			return;
		}
		try
		{
			Process p = Runtime.getRuntime().exec("kde4-config --localprefix");
			BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
			kdeLocalPrefix = input.readLine();
			input.close();
		}
		catch(IOException err)
		{
		}
		if(kdeLocalPrefix == null)
		{
			try
			{
				Process p = Runtime.getRuntime().exec("kde-config --localprefix");
				BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
				kdeLocalPrefix = input.readLine();
				input.close();
			}
			catch(IOException err)
			{
			}
		}
		if(kdeLocalPrefix == null)
		{
			System.out.println("Cannot find kde4-config or kde-config ... assuming default path in " + System.getProperty("user.home") + "/.kde\n");
			kdeLocalPrefix = System.getProperty("user.home") + "/.kde/";
		}
		String digikamrcPath = kdeLocalPrefix + "share/config/digikamrc";
		String dbType = null;
		String dbName = null;
		try
		{
			String line;
			BufferedReader input = new BufferedReader(new FileReader(digikamrcPath));
			while((line = input.readLine()) != null)
			{
				if(line.contains("Database Type"))
				{
					dbType = line.substring(line.indexOf('=') + 1, line.length());
				}
				else if(line.contains("Database Name"))
				{
					dbName = line.substring(line.indexOf('=') + 1, line.length());
				}
			}
		}
		catch(Exception e)
		{
			System.err.println("Cannot read Digikam configuration file at " + digikamrcPath);
			return;
		}
		if(!"QSQLITE".equals(dbType))
		{
			System.err.println("Database Backend is not SQLite, currently unsupported");
			return;
		}
		DigikamDBValidator ddbv = new DigikamDBValidator(dbName + "digikam4.db");
		int errorCount = 0;

		errorCount += ddbv.validateAlbumRoots();
		errorCount += ddbv.validateAlbums();
		errorCount += ddbv.validateImageComments();
		errorCount += ddbv.validateImageCopyright();
		errorCount += ddbv.validateImageHaarMatrix();
		errorCount += ddbv.validateImageInformation();
		errorCount += ddbv.validateImageMetadata();
		errorCount += ddbv.validateImagePositions();
		errorCount += ddbv.validateImageProperties();
		errorCount += ddbv.validateImageTags();
		errorCount += ddbv.validateImages();
		errorCount += ddbv.validateTags();
		errorCount += ddbv.validateTagsTree();
		System.out.println();
		switch (errorCount)
		{
			case 0:
				System.out.println("No error found - Your Digikam database is integer");
				break;
			case 1:
				System.out.println("1 error has been identified in your Digikam database");
				break;
			default:
				System.out.println(errorCount + " errors have been identified in your Digikam database");
				break;
		}
	}
}
